/*******************************************************************************
 * RP6 C-Control PRO M128 Erweiterungsmodul
 * ----------------------------------------------------------------------------
 * I2C Bus - Sensorik - Beispielprogramm
 * ----------------------------------------------------------------------------
 * Die Sensoren des RP6 knnen mit dem I2C Slave Programm komplett ber
 * den Bus ausgelesen werden.
 * Wie dies grundstzlich funktioniert, wird in diesem Programm demonstriert.
 * Es liest die beiden Lichtsensoren vorne am Roboter aus und versucht den
 * Roboter anhand dieser Sensorwerte in die Richtung der hellsten Lichtquelle
 * auszurichten.
 * Man muss evtl. die Werte im Programm anpassen falls es mit den verwendeten
 * Lichtquellen nicht gut genug klappt.
 * (Es kann z.B. vorkommen das der Roboter ein wenig zu weit fhrt, so dass
 * es auf der jeweils anderen Seite wieder heller ist. Dadurch kann es dazu 
 * kommen, dass der Roboter stndig um die korrekte Ausrichtung hin und 
 * herschwingt - in diesem Fall muss man entweder den RANGE Parameter 
 * vergrern oder die Geschwindigkeit ndern )
 *
 * In den folgenden Beispielprogrammen
 * (Example_07_RP6_Sensors1 - Example_07_RP6_Sensors5)
 * werden schrittweise die Kommunikation mit dem RP6 und die Sensorauswertung
 * vervollstndigt. Der Schrittweise Aufbau der Kommunikationsroutinen
 * erleichtert das Verstndnis, sofern man die Programme nacheinander
 * durchgeht und nachvollzieht.
 * Bitte lesen Sie die Kommentare aufmerksam durch!
 *
 * brigens wird gerade bei diesem Programm deutlich, warum auch eine
 * unkommentierte Version (RP6_Sensors_uncommented.cc) vorhanden ist - die
 * unkommentierte Version ist deutlich bersichtlicher wenn man einmal
 * verstanden hat, was die einzelnen Funktionen genau machen.
 * (165 vs. 378 Zeilen Code)
 * ############################################################################
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+
 *
 * ACHTUNG: Der Roboter bewegt sich in diesem Beispielprogramm!
 * Bitte stellen Sie sicher, dass gengend Freiraum um den Roboter
 * vorhanden ist (in diesem Programm dreht er sich nur auf der Stelle, also
 * reichen mind. 20cm Abstand auf allen Seiten)!
 *
 * Achten Sie darauf das Flachbandkabel des USB Interfaces vom Roboter zu
 * trennen wenn sich der Roboter zu weit vom PC entfernen kann.
 *
 * Sie brauchen NICHT die Verbindung des Interfaces zum PC zu trennen! Nur
 * das Flachbandkabel zum Roboter muss entfernt werden!
 * (Das hat den Vorteil, dass der virtuelle USB Port dann nicht abgemeldet
 *  wird und die CCPRO IDE bzw. der RP6Loader den Port nicht neu ffnen mssen)
 * #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+
 * ############################################################################
 *
 ******************************************************************************/

// WICHTIG: Immer die RP6CCLib mit einbinden:
#include "../../RP6CCLib/RP6CCLib.cc"

// I2C Bus Adresse des Controllers auf dem Mainboard:
#define RP6_BASE_ADR 10


/*****************************************************************************/
// Rotations Funktion

byte transmit_buffer[10]; // Temporrer Daten Buffer fr die I2C Kommunikation.
                          // Wird verwendet um Nachrichten die ber den
                          // Bus gesendet werden sollen zusammenzustellen


/**
 * Anm.: Da die folgende Funktion sehr wichtig ist, beschreiben wir sie
 * hier etwas ausfhrlicher. Fr die anderen Bewegungsfunktionen gilt das hier
 * beschriebene natrlich ebenfalls!
 *
 * Die moveAtSpeed Funktion setzt die Motorgeschwindigkeit auf einen
 * bestimmten Wert.
 * Dieser wird dann fortlaufend eingeregelt (d.h. der Roboter versucht
 * automatisch die Geschwindigkeit mglichst konstant zu halten, auch
 * bei wechselnder Belastung).
 *
 * Man muss moveAtSpeed nur dann aufrufen, wenn man etwas an der aktuellen
 * Geschwindigkeit NDERN mchte. Man sollte diese Funktion nicht in einer
 * Endlosschleife ohne Pause aufrufen, da jedesmal ber den I2C Bus
 * ein Datenpaket gesendet wird - und das kostet viel Zeit wenn man
 * das oft hintereinander tut.
 *
 *
 * Die Geschwindigkeit wird in "Encoder Zhlschritten"
 * pro 200ms (= 0.2 Sekunden) eingestellt (s. a. RP6 Anleitung!).
 *
 * Die tatschlich zurckgelegte Strecke pro Zhlschritt variiert von Roboter
 * zu Roboter aufgrund der hohen Auflsung der Drehgeber recht deutlich.
 * Man kann durch abfahren von Teststrecken bestimmter Lnge die tatschliche
 * Auflsung ermitteln, wenn man Wert auf Genauigkeit legt.
 * Wenn die absoluten Werte nicht interessieren, kann man auch einfach von
 * 0.25 oder besser 0.24mm / Zhlschritt ausgehen.
 * (s. a. RP6 Anleitung!).
 *
 * Meist liegt der tatschliche Wert im Bereich von etwa 0.235mm bis 0.25mm pro
 * Zhlschritt. Die Encoder liefern 625 Zhlschritte pro Radumdrehung.
 *
 * Der Controller auf dem Mainboard aktualisiert die gemessenen
 * Geschwindigkeitswerte standardmig alle 200ms = 0.2s also
 * mit einer Frequenz von 5Hz.
 *
 * Die tatschliche Geschwindigkeit errechnet sich bei einer
 * Auflsung von 0.25mm daher wie folgt:
 * speed = encoder_value * 0.25mm * 5 Hz
 *       = encoder_value * 1.25mm *  1/s
 *       = encoder_value * 0.125cm/s.
 * "encoder_value" ist hier die Zahl der Impulse die von den Encodern
 * zurckgeliefert werden.
 * Um aus einer gemessenen Geschwindigkeit in Encoder Zhlschritten
 * also eine Geschwindigkeit in cm/s abzuleiten, muss man einfach den Messwert
 * mit 0.125 multiplizieren (sofern die Encoder Auflsung 0.25mm ist) bzw.
 * durch 8 teilen (0.125 = 1/8). Umgekehrt, kann man durch Multiplikation mit 8
 * auch von cm/s auf den Drehgeber/Encoder Wert umrechnen.
 *
 * Beispiel fr moveAtSpeed: Der Roboter soll mit 10cm/s fahren.
 * Also:
 * 10cm/s = encoder_value * 0.125cm/s
 * => encoder_value = 10 / 0.125 = 80
 *
 * Im Programm schreibt man dann also:
 * moveAtSpeed(80,80); // Mit etwa 10cm/s geradeaus fahren!
 *
 * Weitere Beispiele:
 * moveAtSpeed(40,40); // Mit etwa 5cm/s geradeaus fahren!
 * moveAtSpeed(100,100); // Mit etwa 12.5cm/s geradeaus fahren!
 *
 */
void moveAtSpeed(byte desired_speed_left, byte desired_speed_right)
{
	RP6_writeCMD_2params(RP6_BASE_ADR, CMD_MOVE_AT_SPEED,
                          desired_speed_left, desired_speed_right);
}


// Bewegungsrichtungen:
#define FWD 0
#define BWD 1
#define LEFT 2
#define RIGHT 3

/**
 * Die (dauerhafte) Bewegungsrichtung des Roboters wird mit
 * dieser Funktion gendert. Die Geschwindigkeit und sonstige Dinge werden
 * dabei nicht verndert. Allerdings regelt der Roboter automatisch die
 * Geschwindigkeit herunter bis auf 0 und ndert die Bewegungsrichtung erst
 * dann und beschleunigt wieder auf die zuvor eingestellte Geschwindigkeit.
 * Dies schont die Getriebe und Motoren und vermeidet Strungen der
 * Versorgungsspannung des Roboters (bei schnellem Umschalten der
 * Rotationsrichtung ziehen die Motoren besonders viel Strom)
 * Die nderung der Bewegungsrichtung dauert dadurch natrlich lnger
 * als wenn man diese direkt ndern wrde (was aber bei Bedarf gendert
 * werden kann).
 *
 * Beispiel:
 * changeDirection(LEFT);   // nach links drehen
 * changeDirection(BWD);    // Rckwrts
 * changeDirection(RIGHT);  // Rechts
 * changeDirection(FWD);    // Vorwrts
 */
void changeDirection(byte dir)
{
    RP6_writeCMD_1param(RP6_BASE_ADR, CMD_CHANGE_DIR, dir);
}

/**
 * Die aktuelle Bewegung des Roboters kann man mit dieser Funktion
 * sofort stoppen.
 */
void stop(void)
{
	RP6_writeCMD(RP6_BASE_ADR, CMD_STOP);
}


/*****************************************************************************/
// Sensoren/ADCs:

// Die einzelnen ADC Kanle des Controllers auf dem Mainboard des RP6
// bekommen je eine eigene Variable:
word adcBat;                // Akkuspannung
word adcMotorCurrentLeft;   // Motorstrom Links (Messwert!)
word adcMotorCurrentRight;  // Motorstrom Rechts (Messwert!)
word adcLSL;                // Lichtsensor Links
word adcLSR;                // Lichtsensor Rechts
word adc0;                  // Analog/Digital Wandler Kanal 0
word adc1;                  // Analog/Digital Wandler Kanal 1

// Gemessene Geschwindigkeiten (Messwert! In Zhlschritten pro 1/5 Sekunde):
word mleft_speed;  // Links
word mright_speed; // Rechts

// Zurckgelegte Distanz (Messwert!)
word mleft_dist;  // Links
word mright_dist; // Rechts

// Gewnschte Geschwindigkeit (kein Messwert! Das ist die Sollgeschwindigkeit):
word mleft_des_speed;   // Links
word mright_des_speed;  // Rechts

// Power (automatisch eingeregelter PWM Wert):
word mleft_power;   // Links
word mright_power;  // Rechts

// Buffer fr den Datenempfang
byte sensorBuf[24];

/**
 * Die obigen Variablen aktualisieren sich leider nicht von allein.
 * Man muss dazu diese Funktion aufrufen. Sie liest alle aktuellen Sensorwerte
 * ber den I2C Bus aus und schreibt sie in die obigen Variablen.
 * In der aktuellen Fassung liest diese Funktion 24 Bytes vom Slave
 * Controller.
 *
 * Natrlich dauert das eine gewisse Zeit.
 * Man sollte das also nur so oft wie unbedingt notwendig tun!
 * Bei wichtigen nderungen (Kollision, Hindernis erkannt, Akku leer)
 * wird ausserdem eine Interrupt Leitung gesetzt und man kann
 * Status Register abfragen um zu ermitteln was genau den Interrupt ausgelst
 * hat. In einem anderen Beispielprogramm wird das demonstriert.
 * Im Allgemeinen ist es besser nur die Register auszulesen die man auch
 * wirklich braucht!
 */
void getAllSensors(void)
{
    // 24 Register auslesen - bei Adresse I2C_REG_POWER_LEFT starten...
    // (so werden ein paar Statusregister ausgelassen - dazu kommen wir noch
    // im nchsten Beispielprogramm)
	RP6_readRegisters(RP6_BASE_ADR, I2C_REG_POWER_LEFT, sensorBuf, 24);

	// Die Werte der einzelnen Register befinden sich nun in dem Array
	// sensorBuf und knnen nun den einzelnen Variablen zugewiesen werden.
	mleft_power = sensorBuf[0];
	mright_power = sensorBuf[1];
	mleft_speed = sensorBuf[2];
	mright_speed = sensorBuf[3];
	mleft_des_speed = sensorBuf[4];
	mright_des_speed = sensorBuf[5];
	// Alle weiteren Variablen sind Bit Werte und werden daher jeweils aus
    // zwei einzelnen Registern zusammengesetzt:
	mleft_dist = sensorBuf[6] + (sensorBuf[7]<<8);
	mright_dist = sensorBuf[8] + (sensorBuf[9]<<8);
	adcLSL = sensorBuf[10] + (sensorBuf[11]<<8);
	adcLSR = sensorBuf[12] + (sensorBuf[13]<<8);
	adcMotorCurrentLeft = sensorBuf[14] + (sensorBuf[15]<<8);
	adcMotorCurrentRight = sensorBuf[16] + (sensorBuf[17]<<8);
	adcBat = sensorBuf[18] + (sensorBuf[19]<<8);
	adc0 = sensorBuf[20] + (sensorBuf[21]<<8);
	adc1 = sensorBuf[22] + (sensorBuf[23]<<8);
}

/**
 * Im Gegensatz zur getAllSensors Funktion, liest getLightSensors nur
 * die Werte der zwei Lichtsensoren vom Roboter aus.
 * Das ist sinnvoll wenn man die ganzen anderen Sensorwerte gar nicht
 * braucht!
 * In diesem Programm verwenden wir getAllSensors auch gar nicht,
 * da uns nur die Lichtsensoren interessieren. Aber in den weiteren
 * Beispielprogrammen lagern wir diese Funktion in eine eigene kleine Library
 * aus, daher ist sie hier dennoch schon mit dabei. Die allgemeine Erklrung
 * gilt natrlich auch fr diese Funktion.
 */
void getLightSensors(void)
{
	RP6_readRegisters(RP6_BASE_ADR, I2C_REG_ADC_LSL_L, sensorBuf, 4);
	adcLSL = sensorBuf[0] + (sensorBuf[1]<<8);
	adcLSR = sensorBuf[2] + (sensorBuf[3]<<8);
}

// Maximaler Differenzwert der Lichtsensoren
// bei dem noch nichts unternommen wird:
#define RANGE 25

/**
 * Hier wird demonstriert wie die Lichtsensoren ausgelesen und die Messwerte
 * weiterverarbeitet werden knnen.
 */
void readLightSensors(void)
{
    // Lichtsensoren auslesen:
    getLightSensors();

    // Werte ausgeben:
    newline();
    newline();
    print(" | LSL:"); printInteger(adcLSL);
    print(" | LSR:"); printInteger(adcLSR);
    newline();

    // Werte limitieren um 3 stellig zu bleiben:
    if(adcLSL > 999)
        adcLSL = 999;
    if(adcLSR > 999)
        adcLSR = 999;

    // Linken Sensorwert auf dem LCD ausgeben:
    setCursorPosLCD(1, 2);
	printIntegerLCD(adcLSL);
    setCursorPosLCD(1, 7);

    // Absolute Differenz ermitteln und begrenzen:
    int dif, abs_dif;
    dif = (adcLSL - adcLSR);
    if(dif > 160) { dif = 160; }
    if(dif < -160) { dif = -160; }
    if(dif < 0) { abs_dif = -dif; }
    else
        abs_dif = dif;
    if(abs_dif > 30) abs_dif = 30;

    // Welcher Sensor ist heller? Ist der Unterschied gro genug?
    if(adcLSL < (adcLSR - RANGE)) { // Rechts ist es heller ...
        writeCharLCD('<');
        changeDirection(RIGHT);     // ... also nach rechts drehen!
        moveAtSpeed(abs_dif, abs_dif);
    }
    else if(adcLSR < (adcLSL - RANGE)) { // Links ist es heller ...
        writeCharLCD('>');
        changeDirection(LEFT);           // ... also nach links drehen!
        moveAtSpeed(abs_dif, abs_dif);
    }
    else { // Helligkeitsunterschied zwischen den Sensoren ist klein, also stoppen:
        writeCharLCD('='); stop();
    }

    // Rechten Sensorwert auf dem LCD ausgeben:
    setCursorPosLCD(1, 9);
    printIntegerLCD(adcLSR);

    // Bargraph Display mit den LEDs auf dem Mainboard
    // (zeigt an auf welcher Seite es heller ist):
    if(dif > 140)
        RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SETLEDS, 0x38);
    else if(dif > 75)
        RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SETLEDS, 0x18);
    else if(dif > 35)
        RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SETLEDS, 0x08);
    else if(dif > -35)
        RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SETLEDS, 0x09);
    else if(dif > -75)
        RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SETLEDS, 0x01);
    else if(dif > -140)
        RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SETLEDS, 0x03);
    else
        RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SETLEDS, 0x07);
}


void main(void)
{
	RP6_CCPRO_Init(); // Auf Startsignal warten!

    newline();
    println("   ________________________");
    println("   \\| RP6  ROBOT SYSTEM |/");
    println("    \\_-_-_-_-_-_-_-_-_-_/ ");
    newline();
    println("  RP6 Sensor Reading 1 ");
    newline();

    beep(100, 200, 300);

    // Dieser Text bleibt immer sichtbar, nur die Sensorwerte werden
    // aktualisiert:
    showScreenLCD("LIGHT SENSORS", "L             R");

    // LEDs auf dem Mainboard setzen (die beiden grnen LEDs):
    RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SETLEDS, 0x09);

    while(true)
    {
        // In der Hauptschleife wird stndig die Funktion zur berprfung
        // der Lichtsensoren aufgerufen:
        readLightSensors();
        AbsDelay(10);
    }
}


