/*******************************************************************************
 * RP6 C-Control PRO M128 Erweiterungsmodul
 * ----------------------------------------------------------------------------
 * I2C Bus - Sensorik - Beispielprogramm
 * ----------------------------------------------------------------------------
 * Nun kommt die Interrupt Verarbeitung hinzu. Eigentlich wird das Interrupt
 * Signal hier gar nicht als "echter" Interrupt behandelt. Bei einem
 * Interrupt wrde die Programmausfhrung normalerweise unterbrochen und
 * zu einer speziellen Interrupt Service Routine gesprungen.
 * Das muss man bei der Sensorik des RP6 aber nicht machen. Die zeitkritischen
 * Verarbeitungsschritte fhrt der Controller auf dem Mainboard durch.
 * Auf ein paar Mikrosekunden kommt es bei der Auswertung auf einem
 * Erweiterungsmodul daher nicht an. Was auch gar nicht realisierbar wre,
 * da noch einige Zeit fr die bertragung der Sensordaten ber den I2C Bus
 * bentigt wird.
 *
 * Es wird also anstatt eines Interrupts einfach in einem Thread stndig
 * die Interrupt Leitung vom Slave Controller abgefragt (sog. "Polling").
 * Das berprfen dieses Pins geht sehr schnell und kostet nicht sonderlich
 * viel extra Rechenzeit. Eine Bus Abfrage selbst wrde aber erheblich viel
 * Zeit in Anspruch nehmen. Daher ist die Lsung mit den Interrupt Leitungen
 * sehr zeitsparend.
 *
 * Ist diese Interrupt Leitung gesetzt, ist irgendein Ereignis eingetreten,
 * auf das zum nchstmglichen Zeitpunkt reagiert werden muss.
 * Beispielsweise knnte das ACS ein Hindernis auf der linken Seite detektiert
 * haben. In diesem Fall sollte ein Ausweichmanver eingeleitet werden.
 * Weitere mgliche Signale sind z.B. abgeschlossener Bewegungsablauf,
 * Kollision (Bumper), RC5 Code empfangen oder niedrige Akkuspannung.
 *
 * All das wird in einer neu hinzugefgten Routine thread_checkINT0 berprft.
 * Der Zustand des ACS wird in diesem Beispiel auf dem LCD und mit den LEDs
 * angezeigt. Auf der seriellen Schnittstelle wird zustzlich der Zustand
 * der anderen Sensoren ausgegeben sofern eines der oben genannten Ereignisse
 * eintritt (z.B. wenn einer der Bumper gedrckt wird).
 *
 * Natrlich ist das hier nur ein einfaches Beispiel um das Prinzip zu
 * verdeutlichen. In den nchsten Beispielen wird das weiter verbessert.
 *
 * Die RP6MasterLib.cc wurde auch wieder um ein paar weitere Funktionalitten
 * ergnzt.
 *
 * *****************************************************************************
 * Der Roboter bewegt sich in diesem Beispielprogramm NICHT und kann
 * am Rechner angeschlossen bleiben!
 *
 ******************************************************************************/

#include "../../RP6CCLib/RP6CCLib.cc"

// ###########################
// Die neue Library einbinden - ACHTUNG - diese hat sich in diesem
// Beispielprogramm gendert, da einige neue Funktionalitten hinzugekommen
// sind!
#include "RP6MasterLib.cc"



byte interrupt_status;
#define MASK_DRIVE 	0x80 // Drive System    (1 Bit)
#define MASK_ACS 	0x60 // ACS Status      (2 Bit - Links / Rechts)
#define MASK_BUMPER 0x6 // Bumper Status    (2 Bit - Links / Rechts)
#define MASK_RC5    0x8 // RC5 Reception    (1 Bit)
#define MASK_BATLOW 0x1 // Low Voltage      (1 Bit)

/**
 * Dieser Thread ist fr die berwachung der Interrupt Leitung zustndig und
 * liest bei Bedarf die Statusregister ber den I2C Bus aus.
 */
void thread_checkINT0(void)
{
    while(true) {
        if(Port_ReadBit(PORT_PE5_INT)) { // Hat der Controller auf dem Mainboard neue Daten?

            // In diesem Beispiel werden Interrupt Ereignisse mit einer LED
            // visualisiert.
            Port_WriteBit(PORT_LED2, 1);

            // Die drei Status Register auslesen:
            RP6_readRegisters(RP6_BASE_ADR, 0, messageBuf, 3);


            // Wir berprfen, ob die Status Bits sich verndert haben,
            // also verknpfen wir den alten und neuen Wert per Exklusiv-Oder
            // um nur die genderten Bits auswerten zu mssen.
            byte compare;
            compare = messageBuf[0] ^ interrupt_status;
            interrupt_status = messageBuf[0]; // Update local register


            // ------------------------------------
            // Zunchst prfen wir ob sich die ACS Status Bits verndert haben.
            // Falls nicht, hat irgend etwas anderes den Interrupt ausgelst und
            // wir knnen mit den anderen Abfragen weitermachen.
            if(compare & MASK_ACS)
            {
                 beep_t(200, 1, 1); // Ganz kurz Piepsen!
                 // Die ACS Status Variablen setzen:
                 obstacleLeft = messageBuf[0] & 0x20;
                 obstacleRight = messageBuf[0] & 0x40;

                 byte leds_rp6;
                 leds_rp6 = 0;

                 if(obstacleRight) {  // Hindernis rechts!
                     leds_rp6 = LED3;
                     setCursorPosLCD(1,11);
                     printLCD("RIGHT");
                 }
                 else {               // Kein Hindernis.
                     leds_rp6 = LED1;
                     clearPosLCD(1,11,5) ;
                 }

                 if(obstacleLeft) {  // Hindernis links!
                     leds_rp6 = leds_rp6 | LED6;  // zustzliche LED setzen!
                     setCursorPosLCD(1,0);
                     printLCD("LEFT");
                 }
                 else {               // Kein Hindernis.
                     leds_rp6 = leds_rp6 | LED4;
                     clearPosLCD(1,0,5);
                 }

                 if(obstacleRight && obstacleLeft) {   // Mitte!
                     leds_rp6 = leds_rp6 | LED2 | LED5;
                     setCursorPosLCD(1,6);
                     printLCD("MID");
                 }
                 else
                     clearPosLCD(1,6,3);

                 // LEDs auf dem Mainboard setzen:
                 RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SETLEDS, leds_rp6);
            }

            // ------------------------------------
            // Abfragen ob sich der Zustand der Bumper gendert hat, falls
            // ja, Text auf Display anzeigen und LEDs setzen:
            if(compare & MASK_BUMPER)
            {
                 beep_t(100, 1, 1);
                 bumperLeft = messageBuf[0] & 0x2;
                 bumperRight = messageBuf[0] & 0x4;

                 byte leds;
                 leds = 0;

                 if(bumperRight) {    // Rechter Tastsensor
                     setCursorPosLCD(0,13);
                     printLCD("BR");
                     leds = LED5;
                 }
                 else  // Rechts nichts detektiert - also Text vom LCD lschen:
                     clearPosLCD(0,13,2);

                 if(bumperLeft) {  // Linker Tastsensor
                     setCursorPosLCD(0,0);
                     printLCD("BL");
                     leds = leds | LED3;    // LED zustzlich aktivieren
                 }
                 else
                     clearPosLCD(0,0,2);

                 if(bumperLeft && bumperRight) {  // Beide, also in der Mitte
                     setCursorPosLCD(0,6);
                     printLCD("MID");
                     leds = leds | LED4;
                 }
                 else
                     clearPosLCD(0,6,3);

                 // LEDs auf dem Erweiterungsmodul setzen:
                 setLEDs(leds);
            }

            // Alles weitere wird nur auf der seriellen Schnittstelle ausgegeben,
            // denn die LEDs sind ja schon alle mit ACS und Bumpern belegt ;)
            // Das lsst sich natrlich bei Bedarf ndern!

            // ------------------------------------
            // berprfen ob ein RC5 Datenpaket mit dem IRCOMM empfangen
            // worden ist - falls ja auf der seriellen Schnittstelle ausgeben:
            if(interrupt_status & MASK_RC5)
            {
                beep_t(250, 1, 1);
                RP6_readRegisters(RP6_BASE_ADR, I2C_REG_RC5_ADR, messageBuf, 2);
                print(" RC5:");
                printInteger((messageBuf[0] << 6) | (messageBuf[1]));
                newline();
            }

            // ------------------------------------
            // berprfen ob die Batteriespannung zu niedrig ist, falls ja
            // Warnung ausgeben:
            if(compare & MASK_BATLOW)
            {
                RP6_readRegisters(RP6_BASE_ADR, I2C_REG_ADC_UBAT_L, messageBuf, 2);
    			adcBat = (messageBuf[1] << 8) | (messageBuf[0]);
                print("BAT LOW:");
                printInteger(adcBat);
                newline();
            }

            // ------------------------------------
            // Status des Antriebssystems berprfen:
            if(compare & MASK_DRIVE)
            {
                 print("DRIVE:");
                 printInteger(messageBuf[2]);
                 println(" ");
                 // Bewegung abgeschlossen?
                 movementComplete = messageBuf[2] & 128;
            }
        }
        else
            Port_WriteBit(PORT_LED2, 0);
        Thread_Delay(1);
    }
}


void main(void)
{
	RP6_CCPRO_Init(); // Auf Startsignal warten!

    newline();
    println("   ________________________");
    println("   \\| RP6  ROBOT SYSTEM |/");
    println("    \\_-_-_-_-_-_-_-_-_-_/ ");
    newline();
    println("  RP6 Sensor Reading 3 ");
    newline();

    beep(100, 200, 300);

    showScreenLCD("Sensors", "DEMO Program 3");
    AbsDelay(2000);
    clearLCD();

    // ACS auf mittlere Sendeleistung stellen:
    RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SET_ACS_POWER, ACS_PWR_MED);

    // Der Thread zur berprfung von Statusnderungen wird gestartet:
    Thread_Start(1,thread_checkINT0);

    RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SETLEDS, LED1 | LED4);

    while(true)
    {
        Thread_Delay(10);
    }
}


