/*******************************************************************************
 * RP6 C-Control PRO M128 Erweiterungsmodul
 * ----------------------------------------------------------------------------
 * I2C Bus - Bewegung & Sensorik - Beispielprogramm
 * ----------------------------------------------------------------------------
 * Dieses Beispiel reagiert genau wie das bereits vorangegangene Programm auf die
 * Bumper und stoppt alle Bewegungen wenn der Roboter mit einem Hindernis
 * kollidiert. Allerdings wird nun nicht mehr konstant eine Kreisbewegung
 * durchgefhrt, sondern der Roboter fhrt ein Rechteck ab. Er wechselt dabei
 * nach jeder Runde die Richtung.
 * Dazu wurde ein neuer Thread hinzugefgt.
 *
 * Vor allem die Rotation um 90 ist dabei nicht ganz so einfach wie es
 * zunchst erscheinen mag. Der Roboter kann sich bei der Rotation einzig und
 * allein auf die Daten die die Drehgeber liefern beziehen. Diese mssen aber
 * aufgrund von Rad bzw. Kettenschlupf nicht unbedingt der Realitt entsprechen.
 * Je nach Beschaffenheit des Untergrundes, haben die Ketten mehr oder weniger
 * gute Bodenhaftung. Dadurch fhrt der Roboter meist eine krzere Strecke
 * als gemessen wurde (das muss aber nicht auf jedem Untergrund so sein!).
 * Man kann auch dies nur durch Testreihen und Korrekturwerte verbessern.
 *
 * Ohne gute Kalibrationswerte fr die Encoder (im I2CSlave Programm) kann man
 * also alles von diesem Programm erwarten, aber nicht das es den Roboter
 * genaue Rechtecke fahren lsst ;-)
 * Das ist eine prinzipielle Schwche eines Kettenantriebs und tritt auch
 * bei viel teureren Systemen auf.
 *
 * Mehr dazu lsst sich in der RP6 Anleitung nachlesen.
 ******************************************************************************/

#include "../../RP6CCLib/RP6CCLib.cc"

// ###########################
#include "RP6MasterLib.cc"

// ###########################
#include "RP6_Sensorcheck.cc"


// Bei vielen Threads macht es Sinn, die Thread Nummern durch lesbare
// Namen zu ersetzen:
#define THREAD_CHECK_INT0         1
#define THREAD_CHECK_BUMPERS      2
#define THREAD_BLINK              3
#define THREAD_MOVE_RECT          4

/**
 * Dieser Thread wird ausgefhrt sobald der Roboter mit einem
 * Hindernis kollidiert ist.
 */
void blink(void)
{
    while(true)
    {
        setRP6LEDs(LED2 | LED3);
        setLEDs(LED3 | LED4 | LED5);  // Kurz aufblitzen lassen ...
        Thread_Delay(5);              // ...
        setLEDs(0);                   // ...
        Thread_Delay(30);
        setRP6LEDs(LED5 | LED6);
        setLEDs(LED3 | LED4 | LED5);
        Thread_Delay(5);
        setLEDs(0);
        Thread_Delay(30);
    }
}

/**
 * Thread zum berprfen der Tastsensoren.
 */
void check_bumpers(void)
{
    byte stopped;
    stopped = false;

    while(true)
    {
        if(!stopped) { // Es sollte nur einmal stop() und Thread_Start aufgerufen
                       // werden, daher fragen wir hier ab, ob das bereits
                       // geschehen ist.

            // Man muss die Bumper nicht in RP6_Sensorcheck.cc
            // auswerten. Man kann auch direkt die beiden Variablen
            // der Tastsensoren abfragen:
            if(bumperLeft || bumperRight)
            {
                Thread_Kill(THREAD_MOVE_RECT); // Bewegungs-Thread anhalten
                // ACHTUNG: Der Roboter kann sich danach noch immer
                // weiterbewegen! Die Bewegung muss man noch explizit
                // stoppen:
                stop();


                Thread_Start(THREAD_BLINK,blink);  // Blinken!
                stopped = true;  // Roboter wurde angehalten!
                setCursorPosLCD(1,0);
                printLCD("Ouch! Crash! :-(");
            }
        }
        Thread_Delay(1);
    }
}

// Geschwindigkeiten als Konstanten definieren:
#define MOVE_SPEED 90
#define TURN_SPEED 70
// (so knnen sie leicht zentral gendert werden, das ist vor allem
// bei komplexeren Programmen sehr ntzlich!)

void move_rect(void)
{
    byte move_count;        // Zhler fr Drehbewegungen
    byte turn_direction;    // Mit dieser Variable wird die nderung der
                            // Bewegungsrichtung realisiert.

    move_count = 0;
    turn_direction = LEFT;

    while(true)
    {
        setRP6LEDs(LED3 | LED6);

		// Move 300mm forwards at "MOVE_SPEED":
		RP6_move_mm(MOVE_SPEED, FWD, 300, BLOCKING);

        // Mit den LEDs auf dem Mainboard die Rotationsrichtung anzeigen:
        if(turn_direction==LEFT)
        {
            setRP6LEDs(LED3 | LED4);
        }
        else
        {
            if(turn_direction==RIGHT)
                setRP6LEDs(LED1 | LED6);
        }

		// Um 90 rotieren:
		RP6_rotate(TURN_SPEED, turn_direction, 90, BLOCKING);
		// Das ist meistens nicht 100%ig genau!
		// Abweichungen von + oder - 5 sind normal, selbst mit gut
        // kalibrierten Drehgeberwerten!
		// Es ist abhngig von der Beschaffenheit des Untergrunds und
        // prinzipbedingt bei einem Kettenantrieb. Exakte Drehungen wren
        // nur mit externen Systemen wie einem Kompass oder IR Baken mglich!

		if(move_count >= 3) // Sind wir wieder am Startpunkt angelangt?
		{
            setRP6LEDs(LED1 | LED3 | LED4 | LED6);
			Thread_Delay(100);

            // Mit den LEDs auf dem Mainboard die Rotationsrichtung anzeigen:
            if(turn_direction==LEFT)
            {
                setRP6LEDs(LED3 | LED4);
            }
            else
            {
                if(turn_direction==RIGHT)
                    setRP6LEDs(LED1 | LED6);
            }

			// Nochmal um 90 drehen um korrekte Ausrichtung zu erhalten:
			RP6_rotate(TURN_SPEED, turn_direction, 90, BLOCKING);

			// Bewegungsrichtung ndern:
			if(turn_direction == LEFT)
				turn_direction = RIGHT;
			else
				turn_direction = LEFT;

			// Zhler zurcksetzen:
			move_count = 0;
		}
		else // Wir sind nicht am Startpunkt, also weiterzhlen:
			move_count++;
    }
}

void main(void)
{
	RP6_CCPRO_Init(); // Auf Startsignal warten!

    newline();
    println("   ________________________");
    println("   \\| RP6  ROBOT SYSTEM |/");
    println("    \\_-_-_-_-_-_-_-_-_-_/ ");
    newline();
    println("  RP6 Movement DEMO 4 ");
    newline();

    beep(100, 200, 300);

    showScreenLCD("Movement", "DEMO Program 4");
    AbsDelay(2000);
    clearLCD();

    // ACS auf mittlere Sendeleistung stellen:
    RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SET_ACS_POWER, ACS_PWR_MED);
    RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SET_WDT, true);
    RP6_writeCMD_1param(RP6_BASE_ADR, CMD_SET_WDT_RQ, true);

    // Der Thread zur berprfung von Statusnderungen wird gestartet:
    Thread_Start(THREAD_CHECK_INT0,thread_checkINT0);

    // Thread zur Bumper berprfung starten:
    Thread_Start(THREAD_CHECK_BUMPERS,check_bumpers);

    setCursorPosLCD(1,0);
    printLCD("Moving...    :-)");

    // Im Rechteck bewegen:
    Thread_Start(THREAD_MOVE_RECT,move_rect);

    // Das knnte man natrlich auch in der Hauptschleife erledigen, allerdings
    // msste man dann eine zustzliche Kontrollstruktur zum Stoppen der
    // Bewegung einfhren! In einem extra Thread muss man einfach nur den
    // Thread anhalten.
    while(true)
    {
        Thread_Delay(1);
    }
}


